/*
 * Chess game implementation.
 * By Carlos Varela  <cvarela@uiuc.edu>  February, 1997
 * v0.1
 *
 * v0.2  <cvarela@uiuc.edu>  May, 1998
 *      - Fixed minimax algorithm to create a tree of moves by level.
 *      - Added incomplete chess concepts:  
 *        o castling
 *        o checkmate, draw
 *        o pawn-in-the-way
 *        o pawn-to-last-row
 *        o move repetition implies DRAW
 *      - Reading moves from a file.
 *      - Configuration parameters from command line.
 *
 * v0.3  <cvarela@uiuc.edu>  May, 1998
 *      - Implemented alpha-beta prunning.
 *      - Tried out new heuristic functions.
 *
 * v0.4  <cvarela@uiuc.edu>  May, 1998
 *      - Iterative deepening
 *      - int vs float static values
 *
 * v0.4.3  <cvarela@uiuc.edu>  May, 1998
 *      - Recursive iterative deepening.
 * 
 */

package chess;

import java.util.*;
import java.io.*;

public class Game implements java.io.Serializable {

  /* Piece color constants */
  public static final boolean WHITE = submit.constants.WHITE;
  public static final boolean BLACK = submit.constants.BLACK;
  public static final int INFINITY = submit.constants.INFINITY;  // King's value

  Board board;                  // The game board.
  boolean next = WHITE;         // Who plays next?
  boolean computer = BLACK;     // Color that computer is playing.
  int numberOfMoves = 0;        // Number of moves.
  Vector moves;                 // Vector of moves so far.
  Clock whiteTime, blackTime;   // Clocks for players.
  MoveTree possibleMoves;       // Tree of possible moves.
  Move nextMove;                // Next move to make
  static boolean debug = submit.constants.DEBUG ; // Debugging information
  int levelLookAhead = 4;       // No of levels to look ahead. 
                                // <0 == infinity
  long thinkTime = -1;          // 60 min * 60 sec * 1000 ms / 40 moves =  90000 ms
                                // -1 == forever.
  int levelSearched;            // how many levels did we search?
  Vector debugMoves;            // Vector with (move, score, depth, nodes, time).

  public void play(String args[]){
    try {
      resetGame(args);
      debug("Reset game");
      if (next == Game.WHITE) whiteTime.start();
      else blackTime.start();
      while (true){
	printAscii();
	nextMove = getNextMove();
        doMove(nextMove);
	debugMoves.addElement
	  (new DebugMove(nextMove, 
			 ((possibleMoves == null)? 
			  0 : possibleMoves.getMaxMin()),
			 getLevel(),
			 Algorithm.nodesEvaluated(),
			 (next? blackTime.value() : whiteTime.value())));
      }
    } catch (CheckMateException mate){ // thrown by resetGame or getNextMove
      if (next == mate.getColor()){    // display checkmate
	try {
	  doMove(mate.getMove());
	} catch (DrawException de){} // can not be checkmate and draw!!
	printAscii();
      }
      stopClocks();
      System.out.println(
			 (mate.getColor()? "Black" : "White") + 
			 " has received checkmate!");
    } catch (DrawException draw){  
      // thrown by getNextMove and resetGame 
      //        when there are no possible moves and there is no check.
      // thrown by doMove, when there is move repetition.
      stopClocks();
      System.out.println("Draw!");
    } finally {
      System.out.println(moves+"\n"+debugMoves);
    }
  }

  /**
   * printAscii prints a character version of the game
   */
  public void printAscii(){
    System.out.println("White: " + whiteTime);
    System.out.println("Black: " + blackTime);
    board.printAscii();   
  }

  /**
   * getNextMove gets the next move in this game, either from the
   * computer, or from a human player.
   * chooseBest/validMove throw an exception in case of checkmate or draw.
   */
  Move getNextMove() throws CheckMateException, DrawException{
    Move move;
    if (next == computer) {              // Turn is for computer
      // Using recursive iterative deepening:
      possibleMoves = board.iterativeDeepening(next, thinkTime, levelLookAhead,
					     -Game.INFINITY, Game.INFINITY, 0);
      move = Algorithm.chooseBest(board, next, possibleMoves);
      System.out.println("My move: "+move);
    }
    else {                               // Turn is for human player 
      do {                               // Get a valid move from player
	move = new Move(board);
	debug("Read move: " + move);
      } while (!move.validMove(board, next));
    }
    return move;
  }

  /**
   * doMove receives a Move object and performs the move in this Game.
   * If it is a move repetition, throws a DrawException.
   */
  public void  doMove(Move move) throws DrawException{
    board.doMove(move);
    moves.addElement(move);
    next = !next;                        // Switch next player
    switchClocks();
    if (next == WHITE) 
      numberOfMoves++;                   // Advance move number.   
    if (board.moveRepetition()){
      printAscii();
      throw new DrawException();         // Repetition of moves
    }
  }

  void resetGame(String[] args)
    throws CheckMateException, DrawException{
    board = new Board(this);
    next = WHITE;
    computer = BLACK;
    numberOfMoves = 0;
    moves = new Vector(80);
    debugMoves = new Vector(80);
    whiteTime = new Clock();
    blackTime = new Clock();
    possibleMoves = null;
    try{
      int arg = 0;
      while (arg < args.length){
	if (args[arg].equals("-read")){
	  resetGame(args[++arg]);
	} else if (args[arg].equals("-position")){
	  readPosition(args[++arg]);
	} else if (args[arg].equals("-debug")){
	  debug = true;
	} else if (args[arg].equals("-level")){
	  levelSearched = levelLookAhead = Integer.parseInt(args[++arg]);
	} else if (args[arg].equals("-time")){
	  thinkTime = Long.parseLong(args[++arg]);
	} else if (args[arg].equals("-white")){
	  computer = WHITE;
	} else {
	  System.err.println("Unknown command line argument:"+args[arg]);
	  printUsage();
	}
	arg++;
      }
    } catch (IndexOutOfBoundsException e){
      printUsage();
    } catch (NumberFormatException e){
      printUsage();
    }
  }

  /**
   * resetGame reads a file and starts the game from such position.
   * Chess file has the format:
   * e2-e4 e7-e5
   * f1-g3 d8-f6
   * d2-d4
   * ...
   */
  void resetGame(String file)
    throws CheckMateException, DrawException{
    // Open file
    try {
      MoveReader in = new MoveReader(new BufferedReader(new FileReader(file)));
      while (true){
	// Read a line "e2-e4 e7-e5" or "e2-e4"
	String line = in.readMove();
	// Do appropriate move[s].
	Move m;
	try {
	  m = new Move(board, line);
	} catch (IndexOutOfBoundsException e) {  // Invalid move string, ignore.
	  debug("Ignoring: "+line);
	  continue;
	}
	if (!m.validMove(board, next)){
	  System.err.println("Invalid move "+numberOfMoves+":"+m);
	  System.err.println("No more moves were read from the file");
	  return;
	}
        board.doMove(m);
	moves.addElement(m);
	board.printAscii();
	next = !next;                        // Switch next player
	if (next == WHITE) 
	  numberOfMoves++;                   // Advance move number.
      }
    } catch (FileNotFoundException e) {
      System.err.println(file+" not found.");
      return;
    } catch (EOFException e) {
      return;
//     } catch (IndexOutOfBoundsException e) {  // Invalid move string, ignore.
//       debug("Ignoring: "+line);
//       return;
    } catch (IOException e) {
      e.printStackTrace();
    } catch (NullPointerException e) {
      e.printStackTrace();
      System.exit(0);
    }    
  }

  /**
   * resetGame reads a file and starts the game from such position.
   * Chess position file has the format:
   *
   * r3nrk1/p3qppp/bp2pn2/2pp4/2PP2N1/P1PBPP2/1BQ3PP/1R3RK1 w - - 0 20
   *
   * for a board like:
   * 
   * [--------------
   * r . . . n r k .
   * p . . . q p p p
   * b p . . p n . .
   * . . p p . . . .
   * . . P P . . N .
   * P . P B P P . .
   * . B Q . . . P P
   * . R . . . R K .
   * white to play
   * --------------]
   *
   */
  void readPosition(String file){
    // Open file
    try {
      BufferedReader in = new BufferedReader(new FileReader(file));
      // Read the line 
      // e.g. "r3nrk1/p3qppp/bp2pn2/2pp4/2PP2N1/P1PBPP2/1BQ3PP/1R3RK1 w - - 0 20"
      String line = in.readLine();
      StringTokenizer tokens = new StringTokenizer(line);
      String position = tokens.nextToken();
      next = (tokens.nextToken().charAt(0) == 'w');
      // for now, ignore other tokens (castling, passe, times)?
      board = new Board(this, position);
    } catch (FileNotFoundException e) {
      System.err.println(file+" not found.");
      return;
    } catch (EOFException e) {
      return;
    } catch (IOException e) {
      e.printStackTrace();
    } catch (NullPointerException e) {
      e.printStackTrace();
      System.exit(0);
    }    
  }

  /*
   * Switch clocks after a move has been made.
   */
  void switchClocks(){
    if (next == WHITE){
      blackTime.stop();
      whiteTime.start();
    } else {
      whiteTime.stop();
      blackTime.start();
    }
  }

  /*
   * Stop clocks.
   */
  void stopClocks(){
      blackTime.stop();
      whiteTime.stop();
  }

  boolean computerColor(){
    return computer;
  }

  boolean nextPlayer(){
    return next;
  }

  public void setLevel(int level){
    levelSearched = level;
  }

  int getLevel(){
    return levelSearched;
  }

  public static void debug(Object o){
    if (debug) System.err.println(o);
  }

  void printUsage(){
    System.err.print("Usage:  java chess.Game ");
    System.err.print("[-read <file>] ");
    System.err.print("[-debug] ");
    System.err.print("[-level <l>] ");
    System.err.print("[-time <ms>] ");
    System.err.print("[-white] ");
    System.err.println();
    System.exit(0);
  }

}
